@@ -29,21 +29,14 @@ class View:
 
 
 @dataclass
-class Dataset:
+class ManagedTestCase:
     id: str
-    name: str
+    body: Dict[str, Any]
 
 
 @dataclass
-class DatasetItem:
-    id: str
-    input: str
-    output: str
-
-
-@dataclass
-class DatasetWithItems(Dataset):
-    items: List[DatasetItem]
+class ManagedTestCaseResponse:
+    test_cases: List[ManagedTestCase]
 
 
 @dataclass
