@@ -1,17 +1,14 @@
 import functools
 import random
-from enum import Enum
 from typing import Any
 from typing import Dict
-from typing import Generic
 from typing import List
 from typing import Optional
 from typing import Set
-from typing import TypeVar
 from typing import Union
 
-from autoblocks._impl.prompts.constants import DANGEROUSLY_USE_UNDEPLOYED
-from autoblocks._impl.prompts.constants import UNDEPLOYED
+from autoblocks._impl.prompts.placeholders import TemplatePlaceholder
+from autoblocks._impl.prompts.placeholders import parse_placeholders_from_template
 
 try:
     import pydantic
@@ -31,7 +28,6 @@ class FrozenModel(pydantic.BaseModel):
 class PromptParams(FrozenModel):
     """LLM model parameters for a prompt."""
 
-    version: str
     params: Optional[Dict[str, Any]]
 
 
@@ -39,15 +35,19 @@ class PromptTemplate(FrozenModel):
     """An individual template for a prompt."""
 
     id: str
-    version: str
     template: str
 
+    @functools.cached_property
+    def placeholders(self) -> List[TemplatePlaceholder]:
+        return parse_placeholders_from_template(self.template)
+
 
 class Prompt(FrozenModel):
     """A schema representing a prompt managed in the Autoblocks platform."""
 
     id: str
     version: str
+    revision_id: str = pydantic.Field(..., alias="revisionId")
     params: Optional[PromptParams] = None
     templates: List[PromptTemplate]
 
@@ -56,37 +56,25 @@ class Prompt(FrozenModel):
         """
         Returns the schema expected by our ingestion endpoint's prompt tracking property.
         """
-        return self.model_dump()
+        return self.model_dump(by_alias=True)
 
 
-MinorVersionEnumType = TypeVar("MinorVersionEnumType", bound=Enum)
-
-
-class WeightedMinorVersion(
-    FrozenModel,
-    Generic[MinorVersionEnumType],
-):
-    version: MinorVersionEnumType
+class WeightedMinorVersion(FrozenModel):
+    version: str
     weight: float = pydantic.Field(..., gt=0)
 
-    @property
-    def str_version(self) -> str:
-        return str(self.version.value)
-
 
 class PromptMinorVersion(FrozenModel):
-    # need to do more research here, WeightedMinorVersion requires generics but we don't care what it is in this case,
-    # just that it's an instance of WeightedMinorVersion.
-    version: Union[Enum, List[WeightedMinorVersion]]  # type: ignore
+    version: Union[str, List[WeightedMinorVersion]]
 
     @property
     def str_version(self) -> Optional[str]:
         if isinstance(self.version, list):
             return None
-        return str(self.version.value)
+        return self.version
 
     @property
-    def weighted_versions(self) -> Optional[List[WeightedMinorVersion]]:  # type: ignore
+    def weighted_versions(self) -> Optional[List[WeightedMinorVersion]]:
         if isinstance(self.version, list):
             return self.version
         return None
@@ -96,7 +84,7 @@ class PromptMinorVersion(FrozenModel):
         versions: Set[str] = set()
         if weighted_versions := self.weighted_versions:
             for v in weighted_versions:
-                versions.add(v.str_version)
+                versions.add(v.version)
         elif str_version := self.str_version:
             versions.add(str_version)
         return versions
@@ -106,7 +94,7 @@ class PromptMinorVersion(FrozenModel):
             # Minor version is a weighted list; choose one according
             # to their weights.
             (rand_str_version,) = random.choices(
-                population=[v.str_version for v in weighted_versions],
+                population=[v.version for v in weighted_versions],
                 weights=[v.weight for v in weighted_versions],
                 k=1,
             )
@@ -119,15 +107,11 @@ class PromptMinorVersion(FrozenModel):
 
 
 class AutogeneratePromptConfig(FrozenModel):
-    id: str
-    major_versions: List[str]
+    model_config = pydantic.ConfigDict(coerce_numbers_to_str=True)
 
-    @pydantic.field_validator("major_versions", mode="before")
-    @classmethod
-    def validate_and_transform_versions(cls, raw: List[Any]) -> List[str]:
-        assert isinstance(raw, list), "major_versions must be a list"
-        as_strings = [str(v) for v in raw]
-        return [UNDEPLOYED if v == DANGEROUSLY_USE_UNDEPLOYED else v for v in as_strings]
+    id: str
+    major_version: Optional[str] = None
+    dangerously_use_undeployed_revision: Optional[str] = None
 
 
 class AutogeneratePromptsConfig(FrozenModel):
