@@ -3,7 +3,11 @@ from autoblocks._impl.testing.models import BaseEventEvaluator
 from autoblocks._impl.testing.models import BaseTestCase
 from autoblocks._impl.testing.models import BaseTestEvaluator
 from autoblocks._impl.testing.models import Evaluation
+from autoblocks._impl.testing.models import EvaluationOverride
+from autoblocks._impl.testing.models import EvaluationOverrideComment
+from autoblocks._impl.testing.models import EvaluationOverrideField
 from autoblocks._impl.testing.models import HumanReviewField
+from autoblocks._impl.testing.models import ScoreChoice
 from autoblocks._impl.testing.models import TestCaseConfig
 from autoblocks._impl.testing.models import Threshold
 from autoblocks._impl.testing.models import TracerEvent
@@ -14,8 +18,12 @@ __all__ = [
     "BaseTestCase",
     "BaseTestEvaluator",
     "Evaluation",
+    "EvaluationOverride",
+    "EvaluationOverrideComment",
+    "EvaluationOverrideField",
     "HumanReviewField",
     "TestCaseConfig",
     "Threshold",
     "TracerEvent",
+    "ScoreChoice",
 ]
